/**
 * semsub  --  simple binary semaphore handling subroutines
 *		Kazuro Furukawa, KEK, Japan (furukawa@kek.jp)
 *		(if you modified this code, please inform furukawa)
 *
 *	K.Furukawa, Jan.6.1987. swevent programs for os9
 *	K.F, Aug.14.1993.(2)	semsub for unix and os9
 *	K.F, Sep.15.1993.       fs_reset function
 *	K.F, May.8.1994.        also support posix semaphore
 *	K.F, Aug.27.1998.       os9 error EOS_xxx for __STDC__
 *
 * This package contains simple semaphore handling routines for unix 
 * (system-V), posix and os9.  (on os9 semaphore is called as event.)
 * This module defines these routines.
 *  (int fs_check( key, option );)
 *  int fs_create( key, perm, option );
 *  int fs_attach( key, option );
 *  int fs_detach( key, option );
 *  int fs_release( id, option );
 *  int fs_wait( id, option );
 *  int fs_remove( key, option );
 *  int fs_info( key, info, option );
 *  int fs_reset( id, option );
 *
 * The basic usage below describes wait for resource and release.
 *   fs_create( 123, FS_PALL, FS_OERR );	/* create semaphore
 *   id = fs_attach( 123, FS_OERR );		/* use semaphore
 *   fs_release( id, FS_OERR );			/* release resource
 *   ...
 *   fs_wait( id, FS_OERR );			/* wait resource
 *   ...
 *   fs_release( id, FS_OERR );			/* release resource
 *   ...
 *   fs_detach( 123, FS_OERR );			/* no more use
 *  
 * semsub fs_xxx routines are successfully built and executed on
 * ultrix, sun-os, osf1, hp-ux sony-os, lynx-os and os9-68k.
 * on osf1 and lynx-os, posix may be used.
 *
 * with posix, fs_info is not supported, and int is assumed for 
 * binsemset_t.
 */

char    semsub_what[] =
"@(#)semsub.c v1.2.2 simple binary semaphore, K.Furukawa, Aug.1993-Aug.1998";/*
**/

#include "semsub.h"

/* one of symbols, posix, unix or OS9, must be defined (in semsub.h) */

#ifdef posix					/* posix defined in semsub.h */
#undef unix
#include <binsem.h>				/* binary semaphore */
#define KEYFORM "fs_%d"				/* key format for osk */
static char * fs_keyname();			/* key to name conversion */
static binsemset_t fs_id;			/* internal semaphore id */
#endif /* posix */

#include <stdio.h>				/* for error message */
#include <errno.h>				/* for ENOENT, etc */

#ifdef unix					/* unix defined in semsub.h */
#include <sys/ipc.h>				/* for sem...() */
#include <sys/sem.h>				/* for sem...() */
static int fs_id;				/* internal semaphore id */
#endif /* unix */

#ifdef OSK
#ifdef perror
#undef perror
#endif
#define perror(str) prerr(0,errno);fprintf(stderr,"%s\n",str)
#define KEYFORM "fs_%d"				/* key format for osk */
static char * fs_keyname();			/* key to name conversion */
static int fs_id;				/* internal semaphore id */
#ifndef EOS_EVNF				/* for compatibility */
#define EOS_EVNF   E_EVNF
#define EOS_MEMFUL E_MEMFUL
#endif /* EOS_EVNF */
#endif /* OSK */

static void fs_errmsg();

/**
 *  fs_check
 *	fs_check examines semaphore existence
 *
 *    int fs_check( key, option )
 *
 *	key is the key integer to distinguish from other semaphore
 *	if option is 1, error message is sent to stderr
 *
 *    return:	-2 FS_OTHER	other system failure
 *		-3 FS_EXIST	semaphore with key is already existent
 *		-4 FS_NOEXIST	semaphore with key is non existent
 */
int
 fs_check( key, option )
int key;			/* key value for the semaphore */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_check";			/* name of the routine */

#ifdef posix
    if (-1 == (sem_open(fs_keyname(key),&fs_id)) ) { /* open semaphore */
	if (errno == ENOENT)  {			/* non existent */
	    return FS_NOEXIST;
	} else {				/* unknow failure */
	    if ( option )
		fs_errmsg( 1, nm, "sem_open - semaphore operation failure");
	    return FS_OTHER;
	}
    }
    /* here, we close the semaphore once */
    /* fs_attach may open it again */
    sem_close( fs_id );				/* close semaphore */
#endif /* posix */

#ifdef unix
    if (-1 == (fs_id = semget( key, 1, 0 )))  {	/* get info of semaphore */
	if (errno == ENOENT)  {			/* non existent */
	    return FS_NOEXIST;
	} else {				/* unknow failure */
	    if ( option )
		fs_errmsg( 1, nm, "semget - semaphore operation failure");
	    return FS_OTHER;
	}
    }
#endif /* unix */

#ifdef OSK
    if (-1 == (fs_id = _ev_link(fs_keyname(key)))) { /* link to event */
	if (errno == EOS_EVNF)  {		/* event not found */
	    return FS_NOEXIST;
	} else  {				/* unknown failure */
	    if ( option )
		fs_errmsg(1,nm,"evlink - semaphore attach(link) failure");
	    return FS_OTHER;
	}
    }
    /* here, we decrement the event link count */
    /* real increment/decrement will occur in fs_attach/fs_detach */
    if (-1 == _ev_unlink(fs_id))  {		/* unlink event */
	if ( option )
	    fs_errmsg(1, nm, "ev_unlink - semaphore(event) unlink failure");
	return FS_OTHER;
    }
#endif /* OSK */
    return FS_EXIST;
}

/**
 *  fs_info
 *	fs_info examines information on semaphore
 *
 *    int fs_info( key, info, option )
 *
 *	key is the key integer to distinguish from other semaphore
 *	info is the pointer to four long array, which will receive
 *	event value, permission mode, uid, gid
 *	on OS9 permission mode, uid and gid are are always zero
 *	(meaningless)
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0		success
 *		-2 FS_OTHER	other system failure
 *		-4 FS_NOEXIST	semaphore with key is non existent
 */
int
 fs_info( key, info, option )
int key;			/* key value for the semaphore */
long info[4];			/* information of the memory to be returned */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_info";			/* name of the routine */
    int stat;

    if (FS_EXIST != (stat = fs_check(key, option))) { /* check existence */
	if (stat == FS_NOEXIST)  {		/* non existent */
	    if ( option )
		fs_errmsg(1, nm, "semaphore non existent");
	}
	return stat;
    }

#ifdef posix
    if ( option )
	fs_errmsg( 1, nm, "info - not supported yet with posix");
    return FS_OTHER;
#endif /* posix */

#ifdef unix
    {
	struct semid_ds buf;
	int value;				/* semaphore value */

	if (-1 == semctl(fs_id,0,IPC_STAT,&buf))  { /* get semaphore status */
	    if ( option )
		fs_errmsg( 1, nm, "semctl - semaphore operation failure");
	    return FS_OTHER;
	}
	if (-1 == (value = semctl(fs_id,0,GETVAL,0)))  { /* get value */
	    if ( option )
		fs_errmsg( 1, nm, "semctl - semaphore operation failure");
	    return FS_OTHER;
	}

	info[0] = value;
	info[1] = buf.sem_perm.mode & FS_PALL;
	info[2] = buf.sem_perm.uid;
	info[3] = buf.sem_perm.gid;
    }
#endif /* unix */

#ifdef OSK
    {
	int value;				/* event value */

	if (-1 == (value = _ev_read(fs_id)))  {	/* read event value */
	    if ( option )
		fs_errmsg(1, nm, "ev_read - semaphore(event) read failure");
	    return FS_OTHER;
	}

	info[0] = value;
	info[1] = 0;
	info[2] = 0;
	info[3] = 0;
    }
#endif /* OSK */

    return 0;
}

/**
 *  fs_create
 *	fs_create allocates new semaphore
 *
 *    int fs_create( key, perm, option )
 *
 *	key is the key integer to distinguish from other semaphore
 *	size is the size of the semaphore in byte
 *	perm is the permisson codes (FS_POWNER, FS_PGRUOP or FS_POTHER)
 *	with posix perm does nothing
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  FS_SUCCESS	success
 *		-1 FS_NORES	no memory available
 *		-2 FS_OTHER	other system failure
 *		-3 FS_EXIST	semaphore with key is already existent
 */
int
 fs_create( key, perm, option )
int key;			/* key value for the semaphore */
int perm;			/* permissions */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_create";			/* name of the routine */

    int stat;
    /*
     * check semaphore existence
     */

    if (FS_NOEXIST != (stat = fs_check(key, 0))) {	/* semaphore exist ? */
	if (stat == FS_EXIST)  {			/* existent */
	    if ( option & FS_OINFO )
		fs_errmsg(1,nm,"semaphore existent");
	}
	return stat;					/* yes or error */
    }

    /*
     * if semaphore is non existence, create it
     */

#ifdef posix
    if (-1 == (sem_mksem(fs_keyname(key),1,SEM_UNLOCKED)) )  {	/* create */
	if ( errno == ENOSPC )  {			/* not enough space */
	    if ( option )
		fs_errmsg(1,nm,"sem_mksem - no new semaphore available");
	    return FS_NORES;
	}  else  {
	    if ( option )
		fs_errmsg(1,nm,"sem_mksem - semaphore creation failure");
	    return FS_OTHER;
	}
    }
#endif /* posix */

#ifdef unix
    if (-1 == (fs_id = semget(key,1,IPC_CREAT|perm)))  {
	if ( errno == ENOSPC )  {			/* not enough space */
	    if ( option )
		fs_errmsg(1,nm,"semget - no new semaphore available");
	    return FS_NORES;
	}  else  {
	    if ( option )
		fs_errmsg(1,nm,"semget - semaphore creation failure");
	    return FS_OTHER;
	}
    }
#endif /* unix */

#ifdef OSK
    if (-1 == (int)(fs_id = 
		    _ev_creat(0, -1, 1, fs_keyname(key))))  {
	if ( errno == EOS_MEMFUL )  {			/* not enough space */
	    if ( option )
		fs_errmsg(1,nm,"ev_creat - no new semaphore available");
	    return FS_NORES;
	}  else  {
	    if ( option )
		fs_errmsg(1,nm,
			  "ev_creat - semaphore creation failure");
	    return FS_OTHER;
	}
    }
#endif /* OSK */

    return FS_SUCCESS;
}

/**
 *  fs_attach
 *	fs_attach returns semaphore id to be used with fs_release or fs_wait
 *
 *    int fs_attach( key, option )
 *
 *	key is the key integer to distinguish from other semaphore
 *	if option is 1, error message is sent to stderr
 *
 *    return:	id		success
 *		-2 FS_OTHER	other system failure
 *		-4 FS_NOEXIST	semaphore with key is not existent
 */
int
 fs_attach( key, option )
int key;			/* key value for the semaphore */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_attach";			/* name of the routine */

    int stat;
    /*
     * check semaphore existence
     */
    if (FS_EXIST != (stat = fs_check(key, option))) {	/* semaphore exist ? */
	if (stat == FS_NOEXIST)  {		/* non existent */
	    if ( option )
		fs_errmsg(1,nm,"semaphore non existent");
	}
	return stat;				/* no or error */
    }

#ifdef posix
    /* on posix we should increment the link count */
    if (-1 == (sem_open(fs_keyname(key),&fs_id)) ) { /* open semaphore */
	if ( option )
	    fs_errmsg(1, nm, "sem_open - semaphore open failure");
	return FS_OTHER;
    }
#endif /* posix */

#ifdef OSK
    /* on OSK we should increment the event link count */
    if (-1 == _ev_link(fs_keyname(key)))  {	/* link to event */
	if ( option )
	    fs_errmsg(1, nm, "ev_link - semaphore(event) link failure");
	return FS_OTHER;
    }
#endif /* OSK */

    return (int) fs_id;
}

/**
 *  fs_detach
 *	fs_detach terminate to use the semaphore
 *
 *    int fs_detach( key, option )
 *
 *	key is the key integer to distinguish from other semaphore
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0		success
 *		-2 FS_OTHER	other system failure
 *		-4 FS_NOEXIST	semaphore with key is not existent
 */
int
 fs_detach( key, option )
int key;			/* key value for the semaphore */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_detach";			/* name of the routine */

#ifdef OSK
    int stat;
    /*
     * check semaphore existence
     */
    if (FS_EXIST != (stat = fs_check(key, option))) {	/* semaphore exist ? */
	if (stat == FS_NOEXIST)  {		/* non existent */
	    if ( option )
		fs_errmsg(1,nm,"semaphore non existent");
	}
	return stat;				/* no or error */
    }
    if (-1 == _ev_unlink(fs_id))  {		/* unlink event */
	if ( option )
	    fs_errmsg(1, nm, "ev_unlink - semaphore(event) unlink failure");
	return FS_OTHER;
    }
#endif /* OSK */

    return FS_SUCCESS;
}

/**
 *  fs_release
 *	fs_release releases (signals or posts) semaphore 
 *
 *    int fs_release( id, option )
 *
 *	id is the semaphore id which was returned from fs_attach
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  FS_SUCCESS	success
 *		-2 FS_OTHER	other system failure
 */
int
 fs_release( id, option )
int id;				/* semaphore id */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_release";			/* name of the routine */

#ifdef posix
    if (-1 == sem_post(id, 0))  {		/* release semaphore */
	if ( option )
	    fs_errmsg(1, nm, "sem_post - semaphore operation failure");
	return FS_OTHER;
    }
#endif /* posix */

#ifdef unix
    static struct sembuf sb = { 0, 1, 0 };

    /* sb.sem_num = 0; sb.sem_op = 1; sb.sem_flg = 0; */

    if (-1 == semop(id, &sb, 1))  {		/* release semaphore */
	if ( option )
	    fs_errmsg(1, nm, "semop - semaphore operation failure");
	return FS_OTHER;
    }
#endif /* unix */

#ifdef OSK
    if (-1 == _ev_signal(id, 0))  {		/* declare event */
	if ( option )
	    fs_errmsg(1, nm, "ev_signal - semaphore(event) operation failure");
	return FS_OTHER;
    }
#endif /* OSK */

    return FS_SUCCESS;
}

/**
 *  fs_wait
 *	fs_wait waits for semaphore to be released
 *
 *    int fs_wait( id, option )
 *
 *	id is the semaphore id which was returned from fs_attach
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  FS_SUCCESS	success
 *		-2 FS_OTHER	other system failure
 */
int
 fs_wait( id, option )
int id;				/* semaphore id */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_wait";			/* name of the routine */

#ifdef posix
    if (-1 == sem_wait(id, 0))  {		/* wait semaphore */
	if ( option )
	    fs_errmsg(1, nm, "sem_wait - semaphore operation failure");
	return FS_OTHER;
    }
#endif /* posix */

#ifdef unix
    static struct sembuf sb = { 0, -1, 0 };

    /* sb.sem_num = 0; sb.sem_op = -1; sb.sem_flg = 0; */

    if (-1 == semop(id, &sb, 1))  {		/* wait semaphore */
	if ( option )
	    fs_errmsg(1, nm, "semop - semaphore operation failure");
	return FS_OTHER;
    }
#endif /* unix */

#ifdef OSK
    if (-1 == _ev_wait(id, 1, 1))  {		/* wait event */
	if ( option )
	    fs_errmsg(1, nm, "ev_wait - semaphore(event) operation failure");
	return FS_OTHER;
    }
#endif /* OSK */

    return FS_SUCCESS;
}

/**
 *  fs_reset
 *	fs_reset sets semaphore to initial value (locked state)
 *
 *    int fs_reset( id, option )
 *
 *	id is the semaphore id which was returned from fs_attach
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  FS_SUCCESS	success
 *		-2 FS_OTHER	other system failure
 */
int
 fs_reset( id, option )
int id;				/* semaphore id */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_reset";			/* name of the routine */

#ifdef posix
    if (-1 == sem_ifwait(id, 0))  {		/* lock semaphore */
	if ( option )
	    fs_errmsg(1, nm, "sem_ifwait - semaphore operation failure");
	return FS_OTHER;
    }
#endif /* posix */

#ifdef unix
    if (-1 == semctl(id, 0, SETVAL, 0))  {	/* initialize semaphore */
	if ( option )
	    fs_errmsg(1, nm, "semctl - semaphore operation failure");
	return FS_OTHER;
    }
#endif /* unix */

#ifdef OSK
    if (-1 == _ev_set(id, 0, 0))  {		/* initialize event */
	if ( option )
	    fs_errmsg(1, nm, "ev_set - semaphore(event) operation failure");
	return FS_OTHER;
    }
#endif /* OSK */

    return FS_SUCCESS;
}

/**
 *  fs_remove
 *	fs_remove removes specified semaphore from system
 *
 *    int fs_remove( key, option )
 *
 *	key is the key integer to distinguish from other semaphore
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  FS_SUCCESS	success
 *		-2 FS_OTHER	other system failure
 *		-4 FS_NOEXIST	semaphore with key is not existent
 */
int
 fs_remove( key, option )
int key;			/* key value for the semaphore */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "fs_remove";			/* name of the routine */

    int stat;

    /*
     * check semaphore existence
     */
    if (FS_EXIST != (stat = fs_check(key, option))) {	/* semaphore exist ? */
	if (stat == FS_NOEXIST)  {		/* non existent */
	    if ( option )
		fs_errmsg(1,nm,"semaphore non existent");
	}
	return stat;
    }

    /*
     * remove semaphore
     */
#ifdef posix
    if (-1 == sem_destroy(fs_keyname(key)) )  {	/* remove semaphore */
	if ( option )
	    fs_errmsg(1, nm, "sem_destroy - semaphore operation failure");
	return FS_OTHER;
    }
#endif /* posix */

#ifdef unix
    if (-1 == semctl(fs_id,0,IPC_RMID,0))  {	/* remove semaphore */
	if ( option )
	    fs_errmsg(1, nm, "semctl - semaphore remove failure");
	return FS_OTHER;
    }
#endif /* unix */

#ifdef OSK
    /* event link count is 1, when a event is created */
    /* here decrement link count to be zero */
    if (-1 == _ev_unlink(fs_id))  {		/* unlink event */
	if ( option )
	    fs_errmsg(1, nm, "ev_unlink - semaphore(event) unlink failure");
	return FS_OTHER;
    }

    /* then delete it */
    if (-1 == _ev_delete(fs_keyname(key)))  {	/* remove event */
        if ( option )
            fs_errmsg(1, nm, "ev_delete - semaphore(event) remove failure");
	return FS_OTHER;
    }
#endif /* OSK */

    return FS_SUCCESS;
}

#ifndef unix
/*
 *  fs_keyname
 *      convert key number to keyname string
 */
static char *
 fs_keyname( key )
int key;				/* key */
{
    static char keyname[20];

    sprintf( keyname, KEYFORM, key );		/* convert into keyname */
    return keyname;
}
#endif /* not unix */

/*
 *  fs_errmsg
 *      print error message
 */
static void
 fs_errmsg( flg, nm, msg )
int flg;				/* flag to use perror */
char * nm;				/* routine name where error occur*/
char * msg;				/* error message to display */
{
    extern int errno;
    int errtmp;
    errtmp = errno;
    if ( flg )  perror( nm );
    fprintf( stderr, "%s: %s\n", nm, msg );
    errno = errtmp;
}


#ifdef TEST
/*
 *  fs_test
 *	test program for fs_xxx routines
 */

#define KEY 5432

main(argc, argv)  int argc; char **argv;
{
    int id;
    long lmem[4];
    fs_create( KEY, FS_PALL, FS_OERR );		/* create semaphore */
    printf("fs_create\n");
    id = fs_attach( KEY, FS_OERR );		/* attach semaphore */
    printf("fs_attach\n");
    fs_info( KEY, lmem, FS_OERR );		/* get info. */
    printf("fs_info, value=%ld, perm=%lx(hex), uid=%ld, gid=%ld\n", 
	   lmem[0], lmem[1], lmem[2], lmem[3]);
    fs_release( id, FS_OERR );			/* release */
    printf("fs_release\n");
    fs_info( KEY, lmem, FS_OERR );		/* get info. */
    printf("fs_info, value=%ld, perm=%lx(hex), uid=%ld, gid=%ld\n", 
	   lmem[0], lmem[1], lmem[2], lmem[3]);
    fs_detach( KEY, FS_OERR );			/* detach semaphore */
    printf("fs_detach\n");
    id = fs_attach( KEY, FS_OERR );		/* attach semaphore */
    printf("fs_attach\n");
    fs_info( KEY, lmem, FS_OERR );		/* get info. */
    printf("fs_info, value=%ld, perm=%lx(hex), uid=%ld, gid=%ld\n", 
	   lmem[0], lmem[1], lmem[2], lmem[3]);
    fs_wait( id, FS_OERR );			/* wait */
    printf("fs_wait\n");
    fs_info( KEY, lmem, FS_OERR );		/* get info. */
    printf("fs_info, value=%ld, perm=%lx(hex), uid=%ld, gid=%ld\n", 
	   lmem[0], lmem[1], lmem[2], lmem[3]);
    fs_release( id, FS_OERR );			/* release */
    printf("fs_release\n");
    fs_info( KEY, lmem, FS_OERR );		/* get info. */
    printf("fs_info, value=%ld, perm=%lx(hex), uid=%ld, gid=%ld\n", 
	   lmem[0], lmem[1], lmem[2], lmem[3]);
    fs_wait( id, FS_OERR );			/* wait */
    printf("fs_wait\n");
    fs_info( KEY, lmem, FS_OERR );		/* get info. */
    printf("fs_info, value=%ld, perm=%lx(hex), uid=%ld, gid=%ld\n", 
	   lmem[0], lmem[1], lmem[2], lmem[3]);
    fs_detach( KEY, FS_OERR );			/* detach semaphore */
    printf("fs_detach\n");
    fs_remove( KEY, FS_OERR );			/* remove semaphore */
    printf("fs_remove\n");
    printf("fs_info failure will appear below\n");
    fs_info( KEY, lmem, FS_OERR );		/* get info. */
}

#endif /* TEST */
