/**
 * shmsub  --  shared (common) memory handling subroutines
 *		Kazuro Furukawa, KEK, Japan (furukawa@kek.jp)
 *		(if you modified this code, please inform furukawa)
 *
 *	K.Furukawa, Aug.29.1988. swmodule programs for osk 
 *	K.Furukawa, Jun.3.1992.  shmtest programs for unix
 *	K.F, Apr.27.1993.	 rewritten for os9 as well as unix
 *	K.F, Jul.8.1993.	 make library
 *	K.F, Aug.28.1993.	 add cm_info
 *	K.F, Sep.16.1997.	 add CM_STRICT_SIZE compile time option
 *	K.F, Aug.27.1998.	 os9 error EOS_xxx for __STDC__
 *	K.F, Jun.17.2002.	 support both linux 2.2/2.4 with one library
 *
 * This package contains shared memory handling routines for unix/os9.
 * (on os9 shared memory is called as a data module.)
 * This module defines these routines.
 *  (int cm_check( key, size, option );)
 *  int cm_create( key, size, option );
 *  (void * cm_map_or_create( key, size, option );)
 *  void * cm_map( key, option );
 *  int cm_unmap( desc, option );
 *  int cm_remove( key, option );
 *  int cm_info( key, info, option );
 *
 * The basic usage below describes data exchange between two processes.
 *  process A:
 *   cm_create( 123, 1024, CM_PALL, CM_OERR );	/* create common memory
 *   mem = (int *)cm_map( 123, CM_OERR );	/* map memory
 *   *mem = data;				/* write data
 *   *(mem+1) = data2;
 *   ...
 *   cm_unmap( cm_desc, CM_OERR );		/* unmap memory
 *  process B:
 *   mem = (int *)cm_map( 123, CM_OERR );	/* map memory
 *   data = *mem;				/* read data
 *   data2 = *(mem+1);
 *   ...
 *   cm_unmap( cm_desc, CM_OERR );		/* unmap memory
 *   cm_remove( 123, CM_OERR );			/* remove common memory
 *  
 * shmsub cm_xxx routines are successfully built and executed on
 * ultrix, sun-os, hp-ux, sony-os, lynxos, digital-unix, aix, linux, 
 * and os9-68k. (But not on next-step nor bsdi which are pure BSD.)
 */
/*
 * If CM_STRICT_SIZE is defined at compile time (it's now default), 
 * internal data structure is added on top of the shared memory.  
 * Normally users do not see the structure, since cm_map() returns 
 * the pointer after the added structure. 
 * This structure is added in order to ensure that cm_info() returns 
 * the memory size requested by cm_create().  Without CM_STRICT_SIZE, 
 * the returned size from cm_map() can be larger than the requested 
 * size in cm_create() call, since the kernel may have given us a 
 * larger memory. 
 * The structure is 16 bytes, and first 4-byte integer contains the 
 * originally requested size.  Other 12 bytes are reserved for future 
 * use.
 */
char    shmsub_what[] =
"@(#)shmsub.c v2.5.1 shared memory, K.Furukawa,Kek, Jun.1992-Jun.2002"; /*
**/

#include <stdio.h>				/* for error message */
#include <errno.h>				/* for ENOENT, etc */

#include "shmsub.h"

#ifdef unix					/* unix defined in shmsub.h */
#include <sys/ipc.h>				/* for shm...() */
#include <sys/shm.h>				/* for shm...() */
static int cm_id;				/* internal shared memory id */
#ifdef Lynx
#include <oscalls.h>				/* for shmat() */
#endif /* Lynx */
#endif /* unix */


#ifdef OSK
#define perror(str) prerr(0,errno);fprintf(stderr,"%s\n",str)
#define KEYFORM "cm_%d"				/* key format for osk */
static mod_exec *cm_mod;			/* internal module header */
static char * cm_keyname();			/* key to name conversion */
extern mod_exec *modlink();
extern mod_exec *_mkdata_module();
#ifndef EOS_MNF				/* for compatibility */
#define EOS_MNF    E_MNF
#define EOS_MEMFUL E_MEMFUL
#endif /* EOS_MNF */
#endif /* OSK */

static void cm_errmsg();
int cm_errno;					/* error info for cm_map() */
void * cm_desc;					/* internal descriptor */

/**
 *  cm_check
 *	cm_check examines shared memory existence
 *
 *    int cm_check( key, option )
 *
 *	key is the key integer to distinguish from other shared memory
 *	if option is 1, error message is sent to stderr
 *
 *    return:	-2 CM_OTHER	other system failure
 *		-3 CM_EXIST	memory with key is already existent
 *		-4 CM_NOEXIST	memory with key is non existent
 */
int
 cm_check( key, option )
int key;			/* key value for the shared memory */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "cm_check";			/* name of the routine */

#ifdef unix
    if (-1 == (cm_id = shmget( key, 0, 0 )))  {	/* get info of shared mem */
	if (errno == ENOENT)  {			/* non existent */
	    return CM_NOEXIST;
	} else {				/* unknow failure */
	    if ( option )
		cm_errmsg( 1, nm, "shmget - shared memory operation failure");
	    return CM_OTHER;
	}
    }
#endif /* unix */

#ifdef OSK

    if (-1 == (int)(cm_mod=modlink(cm_keyname(key),0))) { /* link to module */
	if (errno == EOS_MNF)  {		/* module not found */
	    return CM_NOEXIST;
	} else  {				/* unknown failure */
	    if ( option )
		cm_errmsg(1,nm,"modlink - shared memory attach(link) failure");
	    return CM_OTHER;
	}
    }
    if (-1 == munlink(cm_mod))  {		/* unlink module */
	/* here i unlink the module so that this section emulates */
	/* unix shmget call */
	if ( option )
	    cm_errmsg(1, nm, "munlink - shared memory detach(unlink) failure");
    }
#endif /* OSK */
    return CM_EXIST;
}

/**
 *  cm_info
 *	cm_info examines information on shared memory
 *
 *    int cm_info( key, info, option )
 *
 *	key is the key integer to distinguish from other shared memory
 *	info is the pointer to four long array, which will receive
 *	size, permission mode, uid, gid
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  CM_SUCCESS	success
 *		-2 CM_OTHER	other system failure
 *		-4 CM_NOEXIST	memory with key is non existent
 */
int
 cm_info( key, info, option )
int key;			/* key value for the shared memory */
long info[4];			/* information of the memory to be returned */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "cm_info";			/* name of the routine */

#ifdef unix
    /* on linux kernel 2.2/2.4, sizeof (struct shmid_ds) is 48/84 bytes
       this union is a workaround to support both systems */
    union shmid_ds_84 {
	struct shmid_ds buf1;
	unsigned char buf2[84];
    };
    union shmid_ds_84 buf84;
    struct shmid_ds *buf = &buf84;
    int stat;

    if (CM_EXIST != (stat = cm_check(key, option))) { /* check existence */
	if (stat == CM_NOEXIST)  {		/* non existent */
	    if ( option )
		cm_errmsg(1, nm, "shared memory non existent");
	}
	return stat;
    }

    if (-1 == shmctl(cm_id,IPC_STAT,buf))  {
						/* get shared mem status */
	if ( option )
	    cm_errmsg( 1, nm, "shmctl - shared memory operation failure");
	return CM_OTHER;
    }

    info[0] = buf->shm_segsz;
    info[1] = buf->shm_perm.mode & CM_PALL;
    info[2] = buf->shm_perm.uid;
    info[3] = buf->shm_perm.gid;

#endif /* unix */

#ifdef OSK
    if (-1 == (int)(cm_mod=modlink(cm_keyname(key),0)))  { /* link to module */
	if (errno == EOS_MNF)  {		/* module not found */
	    if ( option )
		cm_errmsg(1,nm,"shared memory non existent");
	    return CM_NOEXIST;
	} else  {				/* unknown failure */
	    if ( option )
		cm_errmsg(1,nm,"modlink - shared memory attach(link) failure");
	    return CM_OTHER;
	}
    }

    info[0] = cm_mod->_mh._msize - cm_mod->_mexec;	/* module - header */
    info[1] = cm_mod->_mh._maccess & CM_PALL;
    info[2] = (cm_mod->_mh._mowner) & 0xffff;
    info[3] = ((cm_mod->_mh._mowner)>>16) & 0xffff;

    if (-1 == munlink(cm_mod))  {		/* unlink module */
	if ( option )
	    cm_errmsg(1, nm, "munlink - shared memory detach(unlink) failure");
	return CM_OTHER;
    }
#endif /* OSK */

#ifdef CM_STRICT_SIZE
    {
	int *head;			/* 4-byte integer */
	void *desc;			/* shared memory descriptor */

	desc = cm_desc;				/* save original descriptor */
	if ( !(head = (int *)cm_map( key, option )) )  {
	    return cm_errno;
	}
	head -= CM_INTERN_HEAD/(sizeof (int));
	info[0] = *head;			/* recorded size */
	cm_unmap(cm_desc, option);
	cm_desc = desc;				/* restore descriptor */
    }
#endif /* CM_STRICT_SIZE */

    return CM_SUCCESS;
}

/**
 *  cm_create
 *	cm_create allocate new shared memory
 *
 *    int cm_create( key, size, perm, option )
 *
 *	key is the key integer to distinguish from other shared memory
 *	size is the size of the shared memory in byte
 *	perm is the permisson codes (CM_POWNER, CM_PGRUOP or CM_POTHER)
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  CM_SUCCESS	success
 *		-1 CM_NOMEM	no memory available
 *		-2 CM_OTHER	other system failure
 *		-3 CM_EXIST	memory with key is already existent
 */
int
 cm_create( key, size, perm, option )
int key;			/* key value for the shared memory */
int size;			/* size of the memory to be allocated */
int perm;			/* permissions */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "cm_create";			/* name of the routine */

    int stat;
    /*
     * check shared memory existence
     */

    if (CM_NOEXIST != (stat = cm_check(key, option))) {	/* module exist ? */
	if (stat == CM_EXIST)  {			/* existent */
	    if ( option & CM_OINFO )
		cm_errmsg(1,nm,"shared memory existent");
	}
	return stat;					/* yes or error */
    }

    /*
     * if shared memory is non existence, create it
     */

#ifdef unix
    if (-1 == (cm_id = shmget(key,size+CM_INTERN_HEAD,IPC_CREAT|perm)))  {
	if ( errno == ENOMEM )  {
	    if ( option )
		cm_errmsg(1,nm,"shmget - no new memory available");
	    return CM_NOMEM;
	}  else  {
	    if ( option )
		cm_errmsg(1,nm,"shmget - shared memory creation failure");
	    return CM_OTHER;
	}
    }
#endif /* unix */

#ifdef OSK
    if (-1 == (int)(cm_mod = 
		    _mkdata_module(cm_keyname(key),size+CM_INTERN_HEAD,
				   0x8000|1,perm)))  {
	if ( errno == EOS_MEMFUL )  {
	    if ( option )
		cm_errmsg(1,nm,"shmget - no new memory available");
	    return CM_NOMEM;
	}  else  {
	    if ( option )
		cm_errmsg(1,nm,
			  "mkdata_module - shared memory creation failure");
	    return CM_OTHER;
	}
    }
#endif /* OSK */

#ifdef CM_STRICT_SIZE
    {
	int *head;			/* 4-byte integer */
	void *desc;			/* shared memory descriptor */

	desc = cm_desc;				/* save original descriptor */
	if ( !(head = (int *)cm_map( key, option )) )  {
	    return cm_errno;
	}
	head -= CM_INTERN_HEAD/(sizeof (int));
	*head++ = size;				/* record size */
	*head++ = *head++ = *head++ = 0;
	cm_unmap(cm_desc, option);
	cm_desc = desc;				/* restore descriptor */
    }
#endif /* CM_STRICT_SIZE */

    return CM_SUCCESS;
}

/**
 *  cm_map
 *	cm_map maps shared memory to user address space
 *
 *    void * cm_map( key, option )
 *
 *	key is the key integer to distinguish from other shared memory
 *	if option is 1, error message is sent to stderr
 *
 *    return:	address		success
 *		NULL (void *0)	failure
 *		global integer cm_errno will contain
 *		-2 CM_OTHER	other system failure
 *		-4 CM_NOEXIST	memory with key is non existent
 *		global void * cm_desc will contain shared memory descriptor
 *		which will be required to unmap memory
 */
void *
 cm_map( key, option )
int key;			/* key value for the shared memory */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "cm_map";			/* name of the routine */

    int stat;
    void *addr;
    /*
     * check shared memory existence
     */
    if (CM_EXIST != (stat = cm_check(key, option))) {	/* module exist ? */
	if (stat == CM_NOEXIST)  {		/* non existent */
	    if ( option )
		cm_errmsg(1,nm,"shared memory non existent");
	}
	cm_errno = stat;			/* no or error */
	return NULL;
    }

#ifdef unix
    if (-1 == (int)(addr = shmat(cm_id, 0, 0)))  {  /* attach to shared mem */
	if ( option )
	    cm_errmsg(1, nm, "shmat - shared memory attach failure");
	cm_errno = CM_OTHER;
	return NULL;
    }
    cm_desc = addr;				/* descriptor for unmap */
#endif /* unix */

#ifdef OSK
    if (-1 == (int)(cm_mod=modlink(cm_keyname(key),0))) {  /* link to module */
	if ( option )
	    cm_errmsg(1, nm, "modlink - shared memory attach(link) failure");
	cm_errno = CM_OTHER;
	return NULL;
    }
    addr = ((char *)cm_mod + cm_mod->_mexec);	/* pointer to data area */
    cm_desc = (void *) cm_mod;			/* descriptor for unmap */
#endif /* OSK */

    cm_errno = CM_SUCCESS;

#ifdef CM_STRICT_SIZE
    addr = (char *)addr + CM_INTERN_HEAD;	/* user data after internals */
#endif /* CM_STRICT_SIZE */

    return addr;				/* pointer to user data */
}

/**
 *  cm_unmap
 *	cm_unmap removes shared memory mapping from user address space
 *
 *    int cm_unmap( desc, option )
 *
 *	desc is the descriptor which was return from cm_map in cm_desc
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  CM_SUCCES	success
 *		-2 CM_OTHER	other system failure
 */
int
 cm_unmap( desc, option )
void * desc;			/* shared memory descriptor */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "cm_unmap";			/* name of the routine */

    int stat;

#ifdef unix
    if (-1 == shmdt(desc))  {			/* detach from shared mem */
	if ( option )
	    cm_errmsg(1, nm, "shmdt - shared memory detach failure");
	return CM_OTHER;
    }
#endif /* unix */

#ifdef OSK
    if (-1 == munlink(desc))  {			/* unlink module */
	if ( option )
	    cm_errmsg(1, nm, "munlink - shared memory detach(unlink) failure");
	return CM_OTHER;
    }
#endif /* OSK */

    return CM_SUCCESS;
}

/**
 *  cm_remove
 *	cm_remove removes specified shared memory from system
 *
 *    int cm_remove( key, option )
 *
 *	key is the key integer to distinguish from other shared memory
 *	if option is 1, error message is sent to stderr
 *
 *    return:	0  CM_SUCCES	success
 *		-2 CM_OTHER	other system failure
 */
int
 cm_remove( key, option )
int key;			/* key value for the shared memory */
int option;			/* option 1: error message output */
/*
**/
{
    char *nm = "cm_remove";			/* name of the routine */

    int stat;

    /*
     * check shared memory existence
     */
    if (CM_EXIST != (stat = cm_check(key, option))) {	/* module exist ? */
	if (stat == CM_NOEXIST)  {		/* non existent */
	    if ( option )
		cm_errmsg(1,nm,"shared memory non existent");
	}
	return stat;
    }

    /*
     * remove shared memory
     */
#ifdef unix
    if (-1 == shmctl(cm_id,IPC_RMID,0))  {	/* remove shared mem */
	if ( option )
	    cm_errmsg(1, nm, "shmctl - shared memory remove failure");
	return CM_OTHER;
    }
#endif /* unix */

#ifdef OSK
    if (-1 == munlink(cm_mod))  {		/* remove module */
        if ( option )
            cm_errmsg(1, nm, "munlink - shared memory remove(unlink) failure");
	return CM_OTHER;
    }
    /* here you may repeat munlink until it fails in order to make sure */
    /* some other program may link this module several times */
#endif /* OSK */

    return CM_SUCCESS;
}

#ifdef OSK
/*
 *  cm_keyname
 *      convert key number to keyname string
 */
static char *
 cm_keyname( key )
int key;				/* key */
{
    static char keyname[20];

    sprintf( keyname, KEYFORM, key );		/* convert into keyname */
    return keyname;
}
#endif /* OSK */

/*
 *  cm_errmsg
 *      print error message
 */
static void
 cm_errmsg( flg, nm, msg )
int flg;				/* flag to use perror */
char * nm;				/* routine name where error occur*/
char * msg;				/* error message to display */
{
    extern int errno;
    int errtmp;
    errtmp = errno;
    if ( flg )  perror( nm );
    fprintf( stderr, "%s: %s\n", nm, msg );
    errno = errtmp;
}


#ifdef TEST
/*
 *  cm_test
 *	test program for cm_xxx routines
 */

#define KEY 5432
#define SIZE 321
#define DATA 333
#define DATA2 444

main(argc, argv)  int argc; char **argv;
{
    int *mem;
    long lmem[4];
    cm_create( KEY, (long)SIZE, CM_PALL, CM_OERR ); /* create common memory */
    printf("cm_reate, size=%ld\n", (long)SIZE);
    mem = (int *)cm_map( KEY, CM_OERR );	/* map memory */
    printf("cm_map\n");
    *mem = DATA;				/* write data */
    *(mem+1) = DATA2;
    printf("data written, %d, %d\n", *mem, *(mem+1));
    cm_unmap( cm_desc, CM_OERR );		/* unmap memory */
    printf("cm_unmap\n");
    cm_info( KEY, lmem, CM_OERR );		/* get info. */
    printf("cm_info, size=%ld, perm=%lx(hex), uid=%ld, gid=%ld\n", 
	   lmem[0], lmem[1], lmem[2], lmem[3]);
    mem = (int *)cm_map( KEY, CM_OERR );	/* map memory */
    printf("cm_map\n");
    printf("data read, %d, %d\n", *mem, *(mem+1));
    cm_unmap( cm_desc, CM_OERR );		/* unmap memory */
    printf("cm_unmap\n");
    cm_remove( KEY, CM_OERR );			/* remove memory */
    printf("cm_remove\n");
    printf("cm_info failure will appear below\n");
    cm_info( KEY, lmem, CM_OERR );		/* get info. */
}

#endif /* TEST */
